#include "OV2640.h"
#include <WiFi.h>
#include <WebServer.h>
#include <WiFiClient.h>
#include <ESPmDNS.h>

#include "SimStreamer.h"
#include "OV2640Streamer.h"
#include "CRtspSession.h"

OV2640 cam;
CStreamer *streamer;

WebServer server(80);
WiFiServer rtspServer(8554);

void handle_jpg_stream(void)
{
    WiFiClient client = server.client();
    String response = "HTTP/1.1 200 OK\r\n";
    response += "Content-Type: multipart/x-mixed-replace; boundary=frame\r\n\r\n";
    server.sendContent(response);

    while (1)
    {
        cam.run();
        if (!client.connected())
            break;
        response = "--frame\r\n";
        response += "Content-Type: image/jpeg\r\n\r\n";
        server.sendContent(response);

        client.write((char *)cam.getfb(), cam.getSize());
        server.sendContent("\r\n");
        if (!client.connected())
            break;
    }
}

void handle_jpg(void)
{
    WiFiClient client = server.client();

    cam.run();
    if (!client.connected())
    {
        return;
    }
    String response = "HTTP/1.1 200 OK\r\n";
    response += "Content-disposition: inline; filename=capture.jpg\r\n";
    response += "Content-type: image/jpeg\r\n\r\n";
    server.sendContent(response);
    client.write((char *)cam.getfb(), cam.getSize());
}

void handleNotFound()
{
    String message = "Server is running!\n\n";
    message += "URI: ";
    message += server.uri();
    message += "\nMethod: ";
    message += (server.method() == HTTP_GET) ? "GET" : "POST";
    message += "\nArguments: ";
    message += server.args();
    message += "\n";
    server.send(200, "text/plain", message);
}


void WiFi_Re_Connect() {
  // Previous Network 複数回（3回）チェック
  for ( int i = 0; i < 3; i++ ) {
    int counter = 0;

    WiFi.disconnect(true);  // まずは切断する。True は必須。
    WiFi.mode(WIFI_STA);    // 必須。init wifi mode
    WiFi.begin();

    Serial.print("Connecting to previous network :         ");
    while (WiFi.status() != WL_CONNECTED) {
      delay(500);
      Serial.print(".");
      counter++;
      if (counter >= 10) { //after 5 seconds timeout
        Serial.println();
        counter = 0;
        break;
      }
    }
    if (WiFi.status() == WL_CONNECTED) {
      break;
    }
  }
}


void WiFi_SmartConfig_Connect() {
    int counter = 0;

    Serial.print("Connecting via SmartConfig (Waiting)   : ");
    WiFi.disconnect(true);  // まずは切断。True は必須。
    WiFi.mode(WIFI_AP_STA); // モードが違う。
    delay(500);
    WiFi.beginSmartConfig();
    while (!WiFi.smartConfigDone()) {
      delay(500);
      Serial.print(".");
      counter++;
      if (counter > 1000) {
        Serial.println();
        ESP.restart();      // 1000回（500秒）繰り返してダメなら再起動
      }
    }
    Serial.println();
    Serial.print("Waiting for WiFi : ");
    while (WiFi.status() != WL_CONNECTED) {
      delay(500);
      Serial.print(".");
    }
}

void setup()
{
    Serial.begin(115200);
    while (!Serial)
    {
        ;
    }
    delay(500);
    cam.init(esp32cam_aithinker_config);

    // WiFi 接続 ： SmartConfig で接続した結果が残っている場合、自動で再接続する。
    WiFi_Re_Connect();
    if (WiFi.status() != WL_CONNECTED)
      WiFi_SmartConfig_Connect();
    Serial.println("WiFi Connected.");
    Serial.print("IP Address: ");
    Serial.println(WiFi.localIP());
    // esp32cam.local でアクセスできる。
    delay(500);
    MDNS.begin("esp32cam"); // ホスト名
    delay(500);
    
    server.on("/", HTTP_GET, handle_jpg_stream);
    server.on("/jpg", HTTP_GET, handle_jpg);
    server.onNotFound(handleNotFound);
    server.begin();

    rtspServer.begin();

    streamer = new OV2640Streamer(cam);             // our streamer for UDP/TCP based RTP transport
}

void loop()
{
    server.handleClient();

    uint32_t msecPerFrame = 100;
    static uint32_t lastimage = millis();

    // If we have an active client connection, just service that until gone
    streamer->handleRequests(0); // we don't use a timeout here,
    // instead we send only if we have new enough frames
    uint32_t now = millis();
    if(streamer->anySessions()) {
        if(now > lastimage + msecPerFrame || now < lastimage) { // handle clock rollover
            streamer->streamImage(now);
            lastimage = now;

            // check if we are overrunning our max frame rate
            now = millis();
            if(now > lastimage + msecPerFrame) {
                printf("warning exceeding max frame rate of %d ms\n", now - lastimage);
            }
        }
    }
    
    WiFiClient rtspClient = rtspServer.accept();
    if(rtspClient) {
        Serial.print("client: ");
        Serial.print(rtspClient.remoteIP());
        Serial.println();
        streamer->addSession(rtspClient);
    }
}
